#ifndef s__BSDIPA_CONFIG_H
#define s__BSDIPA_CONFIG_H
#define s_BSDIPA_VERSION "0.9.1"
#define s_BSDIPA_CONTACT "Steffen Nurpmeso <steffen@sdaoden.eu>"
#define s_BSDIPA_32
#undef s_BSDIPA_MAGIC_WINDOW
#undef s_BSDIPA_SMALL
#endif
